@@ -20,6 +20,7 @@
 enum {
     READY,
     UPDATE_DEVICES,
+    AUTHORIZE,
     GOT_PAGE_INFO,
     GOT_LINE,
     SCAN_FAILED,
@@ -35,6 +36,21 @@ typedef struct
     gpointer data;
 } SignalInfo;
 
+typedef struct
+{
+    gchar *device;
+    gchar *source;
+    gint dpi;
+    gchar *scan_mode;
+    gint depth;
+    gboolean multi_page;
+} ScanRequest;
+
+typedef struct
+{
+    gchar *username, *password;
+} Credentials;
+
 typedef enum
 {
     STATE_IDLE = 0,
@@ -47,15 +63,18 @@ typedef enum
 
 struct ScannerPrivate
 {
-    GAsyncQueue *scan_queue;
-    gint dpi;
+    GAsyncQueue *scan_queue, *authorize_queue;
     gboolean running;
-    GThread *thread;
+    GThread *thread, *detect_thread;
 };
 
 G_DEFINE_TYPE (Scanner, scanner, G_TYPE_OBJECT);
 
 
+/* Table of scanner objects for each thread (required for authorization callback) */
+static GHashTable *scanners;
+
+
 static gboolean
 send_signal (SignalInfo *info)
 {
@@ -74,6 +93,12 @@ send_signal (SignalInfo *info)
             g_list_free (devices);
         }
         break;
+    case AUTHORIZE:
+        {
+            gchar *resource = info->data;
+            g_free (resource);
+        }
+        break;
     case GOT_PAGE_INFO:
         {
             ScanPageInfo *page_info = info->data;
@@ -120,6 +145,34 @@ emit_signal (Scanner *scanner, guint sig, gpointer data)
 }
 
 
+static gint get_device_weight (const gchar *device)
+{
+    /* Use locally connected devices first */
+    if (g_str_has_prefix (device, "usb:"))
+       return 0;
+
+    /* Use webcams as a last resort */
+    if (g_str_has_prefix (device, "vfl:"))
+       return 2;
+   
+    return 1;
+}
+
+
+static gint
+compare_devices (ScanDevice *device1, ScanDevice *device2)
+{
+    gint weight1, weight2;
+   
+    weight1 = get_device_weight (device1->name);
+    weight2 = get_device_weight (device2->name);
+    if (weight1 != weight2)
+        return weight1 - weight2;
+   
+    return strcmp (device1->label, device2->label);
+}
+
+
 static void
 poll_for_devices (Scanner *scanner)
 {
@@ -138,29 +191,55 @@ poll_for_devices (Scanner *scanner)
         const SANE_Device *device = *device_iter;
         ScanDevice *scan_device;
         GString *label;
+        gchar *c;
         
         scan_device = g_malloc(sizeof(ScanDevice));
 
         scan_device->name = g_strdup (device->name);
+
         label = g_string_new ("");
-        g_string_printf (label, "%s %s", device->vendor, device->model);
+        /* Abbreviate HP as it is a long string and does not match what is on the physical scanner */
+        if (strcmp (device->vendor, "Hewlett-Packard") == 0)
+            g_string_printf (label, "HP %s", device->model);
+        else
+            g_string_printf (label, "%s %s", device->vendor, device->model);
+        
+        /* Replace underscored in name */
+        for (c = label->str; *c; c++)
+            if (*c == '_')
+                *c = ' ';
+
         scan_device->label = label->str;
         g_string_free (label, FALSE);
 
         devices = g_list_append (devices, scan_device);
     }
 
+    /* Sort devices by priority */
+    devices = g_list_sort (devices, (GCompareFunc) compare_devices);
+
     emit_signal (scanner, UPDATE_DEVICES, devices);
 }
 
 
+static void
+control_option (SANE_Handle handle, SANE_Int index, SANE_Action action, void *value)
+{
+    SANE_Status status;
+    
+    status = sane_control_option (handle, index, action, value, NULL);
+    if (status != SANE_STATUS_GOOD)
+        g_warning ("Error setting control option: %s", sane_strstatus(status));
+}
+
+
 static void
 set_bool_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Bool value)
 {
     SANE_Bool v = value;
     g_return_if_fail (option->type == SANE_TYPE_BOOL);
     g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %s)", option_index, value ? "TRUE" : "FALSE");
-    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v, NULL);
+    control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v);
 }
 
 
@@ -172,33 +251,27 @@ set_int_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_I
     g_return_if_fail (option->type == SANE_TYPE_INT);
 
     if (option->constraint_type == SANE_CONSTRAINT_RANGE) {
-        v *= option->constraint.range->quant;
+        if (option->constraint.range->quant)
+            v *= option->constraint.range->quant;
         if (v < option->constraint.range->min)
             v = option->constraint.range->min;
         if (v > option->constraint.range->max)
             v = option->constraint.range->max;
     }
     g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %d)", option_index, value);
-    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v, NULL);
+    control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v);
 }
 
 
 static void
-set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Fixed value)
+set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, double value)
 {
-    SANE_Fixed v = value;
+    SANE_Fixed v = SANE_FIX (value);
 
     g_return_if_fail (option->type == SANE_TYPE_FIXED);
 
-    if (option->constraint_type == SANE_CONSTRAINT_RANGE) {
-        v *= option->constraint.range->quant;
-        if (v < option->constraint.range->min)
-            v = option->constraint.range->min;
-        if (v > option->constraint.range->max)
-            v = option->constraint.range->max;
-    }
-    g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %d)", option_index, value);
-    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v, NULL);
+    g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %f)", option_index, value);
+    control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v);
 }
 
 
@@ -215,7 +288,7 @@ set_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SAN
     string = g_malloc(sizeof(char) * size);
     strcpy (string, value);
     g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, \"%s\")", option_index, value);
-    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, string, NULL);
+    control_option (handle, option_index, SANE_ACTION_SET_VALUE, string);
     g_free (string);
 }
 
@@ -224,13 +297,18 @@ static void
 log_option (SANE_Int index, const SANE_Option_Descriptor *option)
 {
     GString *string;
-    SANE_String_Const *string_iter;
     SANE_Word i;
+    SANE_Int cap;
     
     string = g_string_new ("");
 
-    g_string_append_printf (string, "Option %d: name='%s', title='%s'",
-                            index, option->name, option->title);
+    g_string_append_printf (string, "Option %d:", index);
+    
+    if (option->name)    
+        g_string_append_printf (string, " name='%s'", option->name);
+    
+    if (option->title)
+        g_string_append_printf (string, " title='%s'", option->title);
 
     switch (option->type) {
     case SANE_TYPE_BOOL:
@@ -255,6 +333,8 @@ log_option (SANE_Int index, const SANE_Option_Descriptor *option)
         g_string_append_printf (string, " type=%d", option->type);
         break;
     }
+    
+    g_string_append_printf (string, " size=%d", option->size);
 
     switch (option->unit) {
     case SANE_UNIT_NONE:
@@ -309,6 +389,42 @@ log_option (SANE_Int index, const SANE_Option_Descriptor *option)
     default:
         break;
     }
+    
+    cap = option->cap;
+    if (cap) {
+        struct {
+            SANE_Int cap;
+            const char *name;
+        } caps[] = {
+            { SANE_CAP_SOFT_SELECT,     "soft-select"},
+            { SANE_CAP_HARD_SELECT,     "hard-select"},
+            { SANE_CAP_SOFT_DETECT,     "soft-detect"},
+            { SANE_CAP_EMULATED,        "emulated"},
+            { SANE_CAP_AUTOMATIC,       "automatic"},
+            { SANE_CAP_INACTIVE,        "inactive"},
+            { SANE_CAP_ADVANCED,        "advanced"},
+            { SANE_CAP_ALWAYS_SETTABLE, "always-settable"},
+            { 0,                        NULL}
+        };
+        int i, n = 0;
+        
+        g_string_append (string, " cap=");
+        for (i = 0; caps[i].cap > 0; i++) {
+            if (cap & caps[i].cap) {
+                cap &= ~caps[i].cap;
+                if (n != 0)
+                    g_string_append (string, ",");
+                g_string_append (string, caps[i].name);
+                n++;
+            }
+        }
+        /* Unknown capabilities */
+        if (cap) {
+            if (n != 0)
+                g_string_append (string, ",");
+            g_string_append_printf (string, "%x", cap);
+        }
+    }
 
     g_debug ("%s", string->str);
     g_string_free (string, TRUE);
@@ -318,23 +434,68 @@ log_option (SANE_Int index, const SANE_Option_Descriptor *option)
 }
 
 
+static gpointer
+detect_thread (Scanner *scanner)
+{
+    while (scanner->priv->running) {
+        poll_for_devices (scanner);
+        g_usleep (1000);
+    }
+    
+    return NULL;
+}
+
+
+static void
+authorization_cb (SANE_String_Const resource, SANE_Char username[SANE_MAX_USERNAME_LEN], SANE_Char password[SANE_MAX_PASSWORD_LEN])
+{
+    Scanner *scanner;
+    Credentials *credentials;
+   
+    scanner = g_hash_table_lookup (scanners, g_thread_self ());
+
+    emit_signal (scanner, AUTHORIZE, g_strdup (resource));
+
+    credentials = g_async_queue_pop (scanner->priv->authorize_queue);
+    strncpy (username, credentials->username, SANE_MAX_USERNAME_LEN);
+    strncpy (password, credentials->password, SANE_MAX_PASSWORD_LEN);
+    g_free (credentials);
+}
+
+
+void
+scanner_authorize (Scanner *scanner, const gchar *username, const gchar *password)
+{
+    Credentials *credentials;
+
+    credentials = g_malloc (sizeof (Credentials));
+    credentials->username = g_strdup (username);
+    credentials->password = g_strdup (password);
+    g_async_queue_push (scanner->priv->authorize_queue, credentials);
+}
+
+
 static gpointer
 scan_thread (Scanner *scanner)
 {
-    gchar *device;
+    ScanRequest *request = NULL;
     SANE_Status status;
     SANE_Handle handle = NULL;
     SANE_Parameters parameters;
     const SANE_Option_Descriptor *option;
     SANE_Int option_index = 0;
     ScanState state = STATE_IDLE;
-    SANE_Int bytes_remaining, line_count = 0, n_read, page_num;
-    SANE_Byte *data;
+    SANE_Int bytes_remaining = 0, line_count = 0, n_read = 0, pass_number = 0, page_number = 0, notified_page = -1;
+    SANE_Byte *data = NULL;
     SANE_Int version_code;
-    gboolean done;
+    gboolean done = FALSE;
+    GError *error = NULL;
+    gchar *open_device = NULL;
+
+    g_hash_table_insert (scanners, g_thread_self (), scanner);
 
     g_debug ("sane_init ()");
-    status = sane_init (&version_code, NULL);
+    status = sane_init (&version_code, authorization_cb);
     if (status != SANE_STATUS_GOOD) {
         g_warning ("Unable to initialize SANE backend: %s", sane_strstatus(status));
         return FALSE;
@@ -344,11 +505,17 @@ scan_thread (Scanner *scanner)
              SANE_VERSION_MINOR(version_code),
              SANE_VERSION_BUILD(version_code));
 
+    // NOTE: This assumes that sane_get_devices is thread safe
+    scanner->priv->detect_thread = g_thread_create ((GThreadFunc) detect_thread, scanner, TRUE, &error);
+    if (error) {
+        g_critical ("Unable to create thread: %s", error->message);
+        g_error_free (error);
+    }
+
     while (scanner->priv->running) {
-        /* Get device to use */
+        /* Look for requests */
         if (state == STATE_IDLE) {
-            GTimeVal timeout = { 1, 0 };
-            device = g_async_queue_timed_pop (scanner->priv->scan_queue, &timeout);
+            request = g_async_queue_pop (scanner->priv->scan_queue);
         } else if (state != STATE_CLOSE) {
             if (g_async_queue_length (scanner->priv->scan_queue) > 0)
                 state = STATE_CLOSE;
@@ -357,14 +524,26 @@ scan_thread (Scanner *scanner)
         /* Interrupted */
         if (!scanner->priv->running)
             break;
-        
+
         switch (state) {
         case STATE_IDLE:
-            if (device == NULL) {
-                poll_for_devices (scanner);
-            } else if (device[0] != '\0') {
-                g_debug ("sane_open ()");
-                status = sane_open (device, &handle);
+             /* Close existing device */
+             if (open_device && (!request->device || strcmp (open_device, request->device) != 0)) {
+                g_debug ("sane_close ()");
+                sane_close (handle);
+                handle = NULL;
+                open_device = NULL;
+            }
+
+            if (request->device) {
+                if (open_device) {
+                    status = SANE_STATUS_GOOD;
+                }
+                else {
+                    g_debug ("sane_open (\"%s\")", request->device);
+                    status = sane_open (request->device, &handle);
+                }
+
                 if (status != SANE_STATUS_GOOD) {
                     g_warning ("Unable to get open device: %s", sane_strstatus (status));
                     emit_signal (scanner, SCAN_FAILED,
@@ -374,10 +553,17 @@ scan_thread (Scanner *scanner)
                     state = STATE_CLOSE;
                 }
                 else {
+                    open_device = g_strdup (request->device);
                     state = STATE_GET_OPTION;
                     option_index = 0;
+                    pass_number = 0;
+                    page_number = 0;
+                    notified_page = -1;
                 }
             }
+            else {
+                state = STATE_CLOSE;
+            }
             break;
 
         case STATE_GET_OPTION:
@@ -389,7 +575,39 @@ scan_thread (Scanner *scanner)
                 log_option (option_index, option);
                 if (option->name) {
                     if (strcmp (option->name, SANE_NAME_SCAN_RESOLUTION) == 0) {
-                        set_fixed_option (handle, option, option_index, scanner->priv->dpi);
+                        if (option->type == SANE_TYPE_FIXED)
+                            set_fixed_option (handle, option, option_index, request->dpi);
+                        else
+                            set_int_option (handle, option, option_index, request->dpi);                            
+                    }
+                    else if (strcmp (option->name, SANE_NAME_SCAN_SOURCE) == 0) {
+                        if (request->source)
+                            set_string_option (handle, option, option_index, request->source);
+                    }
+                    else if (strcmp (option->name, SANE_NAME_BIT_DEPTH) == 0) {
+                        if (request->depth > 0)
+                            set_int_option (handle, option, option_index, request->depth);
+                    }
+                    else if (strcmp (option->name, SANE_NAME_SCAN_MODE) == 0) {
+                        if (request->scan_mode)
+                            set_string_option (handle, option, option_index, request->scan_mode);
+                    }
+
+                    /* Test scanner options (hoping will not effect other scanners...) */
+                    else if (strcmp (option->name, "hand-scanner") == 0) {
+                        set_bool_option (handle, option, option_index, FALSE);
+                    }
+                    else if (strcmp (option->name, "three-pass") == 0) {
+                        set_bool_option (handle, option, option_index, FALSE);
+                    }                    
+                    else if (strcmp (option->name, "test-picture") == 0) {
+                        set_string_option (handle, option, option_index, "Color pattern");
+                    }
+                    else if (strcmp (option->name, "read-delay") == 0) {
+                        set_bool_option (handle, option, option_index, TRUE);
+                    }
+                    else if (strcmp (option->name, "read-delay-duration") == 0) {
+                        set_int_option (handle, option, option_index, 200000);
                     }
                 }
                 option_index++;
@@ -397,18 +615,21 @@ scan_thread (Scanner *scanner)
             break;
             
         case STATE_START:
-            g_debug ("sane_start ()");
+            g_debug ("sane_start (page=%d, pass=%d)", page_number, pass_number);
             status = sane_start (handle);
-            if (status != SANE_STATUS_GOOD) {
+            if (status == SANE_STATUS_GOOD) {
+                state = STATE_GET_PARAMETERS;
+            }
+            else if (status == SANE_STATUS_NO_DOCS) {
+                state = STATE_CLOSE;
+            }
+            else {
                 g_warning ("Unable to start device: %s", sane_strstatus (status));
                 emit_signal (scanner, SCAN_FAILED,
                              g_error_new (SCANNER_TYPE, status,
                                           /* Error display when unable to start scan */
                                           _("Unable to start scan")));
                 state = STATE_CLOSE;
-            } else {
-                page_num = 0;
-                state = STATE_GET_PARAMETERS;
             }
             break;
             
@@ -429,9 +650,12 @@ scan_thread (Scanner *scanner)
                 info->width = parameters.pixels_per_line;
                 info->height = parameters.lines;
                 info->depth = parameters.depth;
+                info->dpi = request->dpi;
 
-                if (page_num == 0)
+                if (page_number != notified_page) {
                     emit_signal (scanner, GOT_PAGE_INFO, info);
+                    notified_page = page_number;
+                }
 
                 /* Prepare for read */
                 bytes_remaining = parameters.bytes_per_line;
@@ -507,26 +731,32 @@ scan_thread (Scanner *scanner)
 
             /* End scan or start next frame */
             if (done) {
-                if (parameters.last_frame)
-                    state = STATE_CLOSE;
+                if (parameters.last_frame) {
+                    pass_number = 0;
+                    if (request->multi_page) {
+                        page_number++;
+                        emit_signal (scanner, IMAGE_DONE, NULL);
+                        state = STATE_START;
+                    }
+                    else
+                        state = STATE_CLOSE;
+                }
                 else {
-                    page_num++;
-                    state = STATE_GET_PARAMETERS;
+                    pass_number++;
+                    state = STATE_START;
                 }
             }
             break;
-            
+
         case STATE_CLOSE:
             emit_signal (scanner, IMAGE_DONE, NULL);
-            if (handle) {
-                g_debug ("sane_close ()");
-                sane_close (handle);
-            }
-            handle = NULL;
             g_free (data);
             data = NULL;
-            g_free (device);
-            device = NULL;
+            g_free (request->device);
+            g_free (request->source);
+            g_free (request->scan_mode);
+            g_free (request);
+            request = NULL;
             state = STATE_IDLE;
             emit_signal (scanner, READY, NULL);            
             break;
@@ -557,17 +787,41 @@ scanner_start (Scanner *scanner)
 
 
 void
-scanner_scan (Scanner *scanner, const char *device, gint dpi_)
+scanner_scan (Scanner *scanner, const char *device, const char *source,
+              gint dpi, ScanMode scan_mode, gint depth, gboolean multi_page)
 {
-    scanner->priv->dpi = dpi_;
-    g_async_queue_push (scanner->priv->scan_queue, g_strdup (device));    
+    ScanRequest *request;
+
+    g_debug ("scanner_scan (\"%s\", %d, %s)", device ? device : "(null)", dpi, multi_page ? "TRUE" : "FALSE");
+    request = g_malloc0 (sizeof (ScanRequest));
+    if (device)
+        request->device = g_strdup (device);
+    if (source)
+        request->source = g_strdup (source);
+    request->dpi = dpi;
+    switch (scan_mode) {
+    case SCAN_MODE_COLOR:
+        request->scan_mode = g_strdup (SANE_VALUE_SCAN_MODE_COLOR);
+        break;
+    case SCAN_MODE_GRAY:
+        request->scan_mode = g_strdup (SANE_VALUE_SCAN_MODE_GRAY);
+        break;
+    case SCAN_MODE_LINEART:
+        request->scan_mode = g_strdup (SANE_VALUE_SCAN_MODE_LINEART);
+        break;
+    default:
+        break;
+    }
+    request->depth = depth;
+    request->multi_page = multi_page;
+    g_async_queue_push (scanner->priv->scan_queue, request);
 }
 
 
 void
 scanner_cancel (Scanner *scanner)
 {
-    g_async_queue_push (scanner->priv->scan_queue, "");
+    scanner_scan (scanner, NULL, NULL, 0, SCAN_MODE_DEFAULT, 0, FALSE);
 }
 
 
@@ -578,6 +832,8 @@ void scanner_free (Scanner *scanner)
     g_async_queue_push (scanner->priv->scan_queue, "");
     if (scanner->priv->thread)
         g_thread_join (scanner->priv->thread);
+    if (scanner->priv->detect_thread)
+        g_thread_join (scanner->priv->detect_thread);    
 
     g_async_queue_unref (scanner->priv->scan_queue);
     g_object_unref (scanner);
@@ -598,6 +854,14 @@ scanner_class_init (ScannerClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__VOID,
                       G_TYPE_NONE, 0);
+    signals[AUTHORIZE] =
+        g_signal_new ("authorize",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (ScannerClass, authorize),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__STRING,
+                      G_TYPE_NONE, 1, G_TYPE_STRING);
     signals[UPDATE_DEVICES] =
         g_signal_new ("update-devices",
                       G_TYPE_FROM_CLASS (klass),
@@ -640,6 +904,8 @@ scanner_class_init (ScannerClass *klass)
                       G_TYPE_NONE, 0);
 
     g_type_class_add_private (klass, sizeof (ScannerPrivate));
+
+    scanners = g_hash_table_new (g_direct_hash, g_direct_equal);
 }
 
 
@@ -649,4 +915,5 @@ scanner_init (Scanner *scanner)
     scanner->priv = G_TYPE_INSTANCE_GET_PRIVATE (scanner, SCANNER_TYPE, ScannerPrivate);
     scanner->priv->running = TRUE;
     scanner->priv->scan_queue = g_async_queue_new ();
+    scanner->priv->authorize_queue = g_async_queue_new ();
 }
